from pysad.core.base_model import BaseModel
from pysad.models.kitnet_model import KitNET as kit


class KitNet(BaseModel):
    """KitNET is a lightweight online anomaly detection algorithm based on an ensemble of autoencoders :cite:`mirsky2018kitsune`. This model directly uses the implementation from `KitNET-py <https://github.com/ymirsky/KitNET-py>`_.

    Args:
        num_features (int): The number of features in your input dataset.
        max_size_ae (int): The maximum size of any autoencoder in the ensemble layer (Default=10).
        grace_feature_mapping (int): The number of instances the network will learn from before producing anomaly scores (Default=None).
        grace_anomaly_detector (int): The number of instances which will be taken to learn the feature mapping. If 'None', then FM_grace_period=AM_grace_period. (Default=50000).
        learning_rate (float): The default stochastic gradient descent learning rate for all autoencoders in the KitNET instance (Default=0.1).
        hidden_ratio (float): The default ratio of hidden to visible neurons. E.g., 0.75 will cause roughly a 25% compression in the hidden layer (Default=0.75).
    """

    def __init__(
            self,
            max_size_ae=10,
            grace_feature_mapping=None,
            grace_anomaly_detector=50000,
            learning_rate=0.1,
            hidden_ratio=0.75):

        self.grace_feature_mapping = grace_feature_mapping
        self.hidden_ratio = hidden_ratio
        self.learning_rate = learning_rate
        self.max_size_ae = max_size_ae
        self.grace_anomaly_detector = grace_anomaly_detector
        self.to_init = True

    def fit_partial(self, X, y=None):
        """Fits the model to next instance. Simply, adds the instance to the window.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to fit.
            y (int): Ignored since the model is unsupervised (Default=None).

        Returns:
            object: Returns the self.
        """
        if self.to_init:
            self.num_features = X.shape[0]
            self.model = kit.KitNET(
                self.num_features,
                self.max_size_ae,
                self.grace_feature_mapping,
                self.grace_anomaly_detector,
                self.learning_rate,
                self.hidden_ratio)
            self.to_init = False
        self.model.process(X)

        return self

    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            float: The anomalousness score of the input instance.
        """
        if self.model.v is None:
            # The feature map is not discovered (i.e., still the grace period),
            # thus, KitNet gives an error.
            return 0.0
        else:
            return self.model.execute(X)



"""
The :mod:`pysad.utils` module includes utility functions used in the `PySAD` framework, which can also be useful streaming learning.
"""
from .array_streamer import ArrayStreamer
import random
import numpy as np
from .data import Data
from .pandas_streamer import PandasStreamer
from .window import Window

__all__ = ["fix_seed", "get_minmax_array", "get_minmax_scalar", "_iterate", "ArrayStreamer", "PandasStreamer", "Window", "Data"]


def fix_seed(seed):
    """Utility method to fix the seed for randomness.

    Args:
        seed (int): The seed.
    """
    random.seed(seed)
    np.random.seed(seed)


def get_minmax_array(X):
    """Utility method that returns the boundaries for each feature of the input array.

    Args:
        X (np.float64 array of shape (num_instances, num_features)): The input array.
    Returns:
        min (np.float64 array of shape (num_features,)): Minimum values for each feature in array.
        max (np.float64 array of shape (num_features,)): Maximum values for each feature in array.
    """
    min = np.min(X, axis=0)
    max = np.max(X, axis=0)

    return min, max


def get_minmax_scalar(x):
    """Utility method that returns the boundaries of the input array.

    Args:
        X (np.float64 array of any shape): Input array.
    Returns:
        float: Minimum value in array.
        float: Maximum value in array.
    """
    min = np.min(x)
    max = np.max(x)

    return min, max


def _iterate(X, y=None):
    """Iterates array of features and possibly labels.

    Args:
        X (np.array of shape (num_instances, num_features)): The features array.
        y (np.array of shape (num_instances, ): The array containing labels (Default=None).
    """

    iterator = ArrayStreamer(shuffle=False)

    if y is None:
        for xi in iterator.iter(X):
            yield xi, None
    else:
        for xi, yi in iterator.iter(X, y):
            yield xi, yi



from abc import abstractmethod
from sklearn.random_projection import SparseRandomProjection, GaussianRandomProjection
from pysad.core.base_transformer import BaseTransformer


class BaseSKLearnProjector(BaseTransformer):

    def __init__(self, num_components):
        """Abstract base projector class to wrap the random sklearn projectors.

        Args:
            num_components (int): The number of dimensions that the target will be projected into.
        """
        super().__init__(num_components)

    @property
    @abstractmethod
    def _projector(self):
        """ Helper property to wrap sklearn projectors.

        """
        pass

    def fit_partial(self, X):
        """Fits particular (next) timestep's features to train the projector.

        Args:
            X (np.float64 array of shape (num_components,)): Input feature vector.
        Returns:
            object: self.
        """
        return self

    def transform_partial(self, X):
        """Projects particular (next) timestep's vector to (possibly) lower dimensional space.

        Args:
            X (np.float64 array of shape (num_features,)): Input feature vector.

        Returns:
            projected_X: np.float64 array of shape (num_components,)
                Projected feature vector.
        """
        x = X.reshape(1, -1)

        return self._projector().fit_transform(x).reshape(-1)


class GaussianRandomProjector(BaseSKLearnProjector):
    """Reduces dimensionality through Gaussian random projection. The components of the random matrix are drawn from N(0, 1 / n_components). This text is taken from the `Sklearn documentation <https://scikit-learn.org/stable/modules/generated/sklearn.random_projection.GaussianRandomProjection.html#sklearn.random_projection.GaussianRandomProjection>`_.

    Args:
        n_components (int or 'auto'): Dimensionality of the target projection space, optional (default = 'auto').

            n_components can be automatically adjusted according to the
            number of samples in the dataset and the bound given by the
            Johnson-Lindenstrauss lemma. In that case the quality of the
            embedding is controlled by the ``eps`` parameter.

            It should be noted that Johnson-Lindenstrauss lemma can yield
            very conservative estimated of the required number of components
            as it makes no assumption on the structure of the dataset.

        eps (strictly positive float, optional): (default=0.1)
            Parameter to control the quality of the embedding according to
            the Johnson-Lindenstrauss lemma when n_components is set to
            'auto'.

            Smaller values lead to better embedding and higher number of
            dimensions (n_components) in the target projection space.

    """

    def __init__(self, num_components='auto', *, eps=0.1):
        super().__init__(num_components)
        self.eps = eps
        self.num_components = num_components

    def _projector(self):
        return GaussianRandomProjection(
            n_components=self.num_components, eps=self.eps)


class SparseRandomProjector(BaseSKLearnProjector):
    """The wrapper method for Sklearn's SparseRandomProjection. Reduces dimensionality through Gaussian random projection. The components of the random matrix are drawn from N(0, 1 / n_components). This text is taken from the `Sklearn documentation <https://scikit-learn.org/stable/modules/generated/sklearn.random_projection.SparseRandomProjection.html#sklearn.random_projection.SparseRandomProjection>`_.

    Parameters
        n_components (int or 'auto'): Optional (default = 'auto')
            Dimensionality of the target projection space.

            n_components can be automatically adjusted according to the
            number of samples in the dataset and the bound given by the
            Johnson-Lindenstrauss lemma. In that case the quality of the
            embedding is controlled by the ``eps`` parameter.

            It should be noted that Johnson-Lindenstrauss lemma can yield
            very conservative estimated of the required number of components
            as it makes no assumption on the structure of the dataset.

        eps (strictly positive float): Optional (default=0.1)
            Parameter to control the quality of the embedding according to
            the Johnson-Lindenstrauss lemma when n_components is set to
            'auto'.

            Smaller values lead to better embedding and higher number of
            dimensions (n_components) in the target projection space.

    """

    def __init__(self, num_components='auto', density="auto", eps=0.1):
        super().__init__(num_components)
        self.eps = eps
        self.density = density
        self.num_components = num_components

    def _projector(self):
        return SparseRandomProjection(
            n_components=self.num_components,
            density=self.density,
            eps=self.eps,
            dense_output=True)



import math
from pysad.core.base_postprocessor import BasePostprocessor
from pysad.statistics.average_meter import AverageMeter
from pysad.statistics.running_statistic import RunningStatistic
from pysad.statistics.variance_meter import VarianceMeter
import numpy as np


class GaussianTailProbabilityCalibrator(BasePostprocessor):
    """Assuming that the scores follow normal distribution, this class provides an interface to convert the scores into probabilities via Q-function, i.e., the tail function of Gaussian distribution :cite:`ahmad2017unsupervised`.

        Args:
            running_statistics (bool): Whether to calculate the mean and variance through running window. The window size is defined by the `window_size` parameter.
            window_size (int): The size of window for running average and std. Ignored if `running_statistics` parameter is False.
    """

    def __init__(self, running_statistics=True, window_size=6400):
        self.running_statistics = running_statistics
        self.window_size = window_size

        if self.running_statistics:
            self.avg_meter = RunningStatistic(AverageMeter, self.window_size)
            self.var_meter = RunningStatistic(VarianceMeter, self.window_size)
        else:
            self.avg_meter = AverageMeter()
            self.var_meter = RunningStatistic(VarianceMeter, self.window_size)

    def fit_partial(self, score):
        """Fits particular (next) timestep's score to train the postprocessor.

        Args:
            score (float): Input score.
        Returns:
            object: self.
        """
        self.avg_meter.update(score)
        self.var_meter.update(score)

        return self

    def transform_partial(self, score):
        """Transforms given score.

        Args:
            score (float): Input score.

        Returns:
            float: Processed score.
        """
        mean = self.avg_meter.get()
        var = self.var_meter.get()
        if var > 0:
            std = np.sqrt(var)
        else:
            std = 1.0

        return 1 - self._qfunction(score, mean, std)

    def _qfunction(self, x, mean, std):
        """
        Given the normal distribution specified by the mean and standard deviation args, return the probability of getting samples > x. Implementation is adapted from the https://github.com/ish-vlad/Conformal-Anomaly-Detection/blob/22769b8d3cede7fabd978a36cdd2853255e450ac/scripts/nab_module/nab/detectors/gaussian/windowedGaussian_detector.py This is the
        Q-function: the tail probability of the normal distribution.
        """

        # Calculate the Q function with the complementary error function, explained
        # here:
        # http://www.gaussianwaves.com/2012/07/q-function-and-error-functions
        z = (x - mean) / std
        return 0.5 * math.erfc(z / math.sqrt(2))





def test_instance_unit_norm_scaler():
    import numpy as np
    from pysad.transform.preprocessing import InstanceUnitNormScaler

    X = np.random.rand(100, 25)
    scaler = InstanceUnitNormScaler()

    scaled_X = scaler.fit_transform(X)
    assert np.all(np.isclose(np.linalg.norm(scaled_X, axis=1), 1.0))

    scaler = scaler.fit(X)
    scaled_X = scaler.transform(X)
    assert np.all(np.isclose(np.linalg.norm(scaled_X, axis=1), 1.0))



from pysad.core.base_postprocessor import BasePostprocessor
from pysad.statistics.average_meter import AverageMeter
from pysad.statistics.max_meter import MaxMeter
from pysad.statistics.median_meter import MedianMeter
from pysad.statistics.variance_meter import VarianceMeter
import numpy as np


class AveragePostprocessor(BasePostprocessor):
    """A postprocessor that convert a score to the average of of all previous scores.
    """

    def __init__(self):
        self.meter = AverageMeter()

    def fit_partial(self, score):
        """Fits the postprocessor to the (next) timestep's score.

        Args:
            score (float): Input score.

        Returns:
            object: self.
        """

        self.meter.update(score)

        return self

    def transform_partial(self, score=None):
        """Gets the current average. This method should be used immediately after the fit_partial method with same score.

        Args:
            score (float): The input score.

        Returns:
            float: Transformed score.
        """
        return self.meter.get()


class MaxPostprocessor(BasePostprocessor):
    """A postprocessor that convert a score to the maximum of of all previous scores.
    """

    def __init__(self):
        self.meter = MaxMeter()

    def fit_partial(self, score):
        """Fits the postprocessor to the (next) timestep's score.

        Args:
            score (float): Input score.

        Returns:
            object: self.
        """
        self.meter.update(score)

        return self

    def transform_partial(self, score=None):
        """Applies postprocessing to the score. This method should be used immediately after the fit_partial method with same score.

        Args:
            score (float): The input score.

        Returns:
            float: Transformed score.
        """
        return self.meter.get()


class MedianPostprocessor(BasePostprocessor):
    """A postprocessor that convert a score to the median of of all previous scores.
    """

    def __init__(self):
        self.meter = MedianMeter()

    def fit_partial(self, score):
        """Fits the postprocessor to the (next) timestep's score.

        Args:
            score (float): Input score.

        Returns:
            object: self.
        """
        self.meter.update(score)

        return self

    def transform_partial(self, score=None):
        """Applies postprocessing to the score.

        Args:
            score (float): The input score.

        Returns:
            float: Transformed score.
        """
        return self.meter.get()


class ZScorePostprocessor(BasePostprocessor):
    """A postprocessor that normalize the score via Z-score normalization.
    """

    def __init__(self):
        self.variance_meter = VarianceMeter()
        self.average_meter = AverageMeter()

    def fit_partial(self, score):
        """Fits the postprocessor to the (next) timestep's score.

        Args:
            score (float): Input score.

        Returns:
            object: self.
        """
        self.variance_meter.update(score)
        self.average_meter.update(score)

        return self

    def transform_partial(self, score):
        """Applies postprocessing to the score.

        Args:
            score (float): The input score.

        Returns:
            float: Transformed score.
        """
        zscore = (score - self.average_meter.get()) / \
            np.sqrt(self.variance_meter.get())

        return zscore



from pysad.core.base_model import BaseModel
from pysad.statistics.average_meter import AverageMeter
from pysad.statistics.median_meter import MedianMeter
from pysad.statistics.variance_meter import VarianceMeter


class StandardAbsoluteDeviation(BaseModel):
    """The model that assigns the deviation from the mean (or median) and divides with the standard deviation. This model is based on the 3-Sigma rule described in :cite:`hochenbaum2017automatic`.

        substracted_statistic (str): The statistic to be substracted for scoring. It is either "mean" or "median". (Default="mean").
        absolute (bool): Whether to output score's absolute value. (Default=True).
    """

    def __init__(self, substracted_statistic="mean", absolute=True):
        self.absolute = absolute
        self.variance_meter = VarianceMeter()

        if substracted_statistic == "median":
            self.sub_meter = MedianMeter()
        elif substracted_statistic == "mean":
            self.sub_meter = AverageMeter()
        else:
            raise ValueError(
                "Unknown substracted_statistic value! Please choose median or mean.")

    def fit_partial(self, X, y=None):
        """Fits the model to next instance.

        Args:
            X (np.float64 array of shape (1,)): The instance to fit. Note that this model is univariate.
            y (int): Ignored since the model is unsupervised (Default=None).

        Returns:
            object: Returns the self.
        """
        assert len(X) == 1  # Only for time series

        self.variance_meter.update(X)
        self.sub_meter.update(X)

        return self

    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (np.float64 array of shape (1,)): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            float: The anomalousness score of the input instance.
        """
        sub = self.sub_meter.get()
        dev = self.variance_meter.get()**0.5

        score = (X - sub) / (dev + 1e-10)

        return abs(score) if self.absolute else score



from pysad.utils.array_streamer import ArrayStreamer
from pysad.core.base_streamer import BaseStreamer


class PandasStreamer(BaseStreamer):
    """Simulator class to iterate dataframe(s).

    Args:
        shuffle (bool): Whether shuffle the data initially (Default=False).
    """

    def __init__(self, shuffle=False):
        super().__init__(shuffle=shuffle)

        self.array_iterator = ArrayStreamer(shuffle=shuffle)

    def iter(self, X, y=None):
        """Iterates pandas dataframes of of features and possibly labels.

        Args:
            X: Pandas Dataframe for features.
            y: Pandas dataframe for labels.
        """
        if y is None:
            for x in self.array_iterator.iter(X.to_numpy()):
                yield x
        else:
            assert len(X) == len(y)

            for x, yr in self.array_iterator.iter(X.to_numpy(), y.to_numpy()):
                yield x, yr



"""
The :mod:`pysad.models.integrations` module contains models to integrate batch anomaly detection models to the streaming setting.
"""
from .one_fit_model import OneFitModel
from .pyod_model import PYODModel
from .reference_window_model import ReferenceWindowModel

__all__ = ["PYODModel", "OneFitModel", "ReferenceWindowModel"]



from pysad.core.base_model import BaseModel
import numpy as np


class RSHash(BaseModel):
    """Subspace outlier detection in linear time with randomized hashing :cite:`sathe2016subspace`. This implementation is adapted from `cmuxstream-baselines <https://github.com/cmuxstream/cmuxstream-baselines/blob/master/Dynamic/RS_Hash/sparse_stream_RSHash.py>`_.

        Args:
            feature_mins (np.float64 array of shape (num_features,)): Minimum boundary of the features.
            feature_maxes (np.float64 array of shape (num_features,)): Maximum boundary of the features.
            sampling_points (int): The number of sampling points (Default=1000).
            decay (float): The decay hyperparameter (Default=0.015).
            num_components (int): The number of ensemble components (Default=100).
            num_hash_fns (int): The number of hashing functions (Default=1).
    """

    def __init__(
            self,
            feature_mins,
            feature_maxes,
            sampling_points=1000,
            decay=0.015,
            num_components=100,
            num_hash_fns=1):
        self.minimum = feature_mins
        self.maximum = feature_maxes

        self.m = num_components
        self.w = num_hash_fns
        self.s = sampling_points
        self.dim = len(self.minimum)
        self.decay = decay
        self.scores = []
        self.num_hash = num_hash_fns
        self.cmsketches = []
        self.effS = max(1000, 1.0 / (1 - np.power(2, -self.decay)))

        self.f = np.random.uniform(
            low=1.0 / np.sqrt(self.effS), high=1 - (1.0 / np.sqrt(self.effS)), size=self.m)

        for i in range(self.num_hash):
            self.cmsketches.append({})

        self._sample_dims()

        self.alpha = self._sample_shifts()

        self.index = 0 + 1 - self.s

        self.last_score = None

    def fit_partial(self, X, y=None):
        """Fits the model to next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to fit.
            y (int): Ignored since the model is unsupervised (Default=None).

        Returns:
            object: Returns the self.
        """
        score_instance = 0
        for r in range(self.m):
            Y = -1 * np.ones(len(self.V[r]))
            Y[range(len(self.V[r]))] = np.floor(
                (X[np.array(self.V[r])] + np.array(self.alpha[r])) / float(self.f[r]))

            mod_entry = np.insert(Y, 0, r)
            mod_entry = tuple(mod_entry.astype(np.int32))

            c = []
            for w in range(len(self.cmsketches)):
                try:
                    value = self.cmsketches[w][mod_entry]
                except KeyError:
                    value = (self.index, 0)

                # Scoring the Instance
                tstamp = value[0]
                wt = value[1]
                new_wt = wt * np.power(2, -self.decay * (self.index - tstamp))
                c.append(new_wt)

                # Update the instance
                new_tstamp = self.index
                self.cmsketches[w][mod_entry] = (new_tstamp, new_wt + 1)

            min_c = min(c)
            c = np.log(1 + min_c)
            score_instance = score_instance + c

        self.last_score = score_instance / self.m

        self.index += 1

        return self

    def score_partial(self, X):
        """Scores the anomalousness of the next instance. Outputs the last score. Note that this method must be called after the fit_partial

        Args:
            X (any): Ignored.
        Returns:
            float: The anomalousness score of the last fitted instance.
        """
        return self.last_score

    def _sample_shifts(self):
        alpha = []
        for r in range(self.m):
            alpha.append(
                np.random.uniform(
                    low=0,
                    high=self.f[r],
                    size=len(self.V[r])))

        return alpha

    def _sample_dims(self):
        max_term = np.max((2 * np.ones(self.f.size), list(1.0 / self.f)), axis=0)
        common_term = np.log(self.effS) / np.log(max_term)
        low_value = 1 + 0.5 * common_term
        high_value = common_term

        self.r = np.empty([self.m, ], dtype=int)
        self.V = []
        for i in range(self.m):
            if np.floor(low_value[i]) == np.floor(high_value[i]):
                self.r[i] = 1
            else:
                self.r[i] = min(
                    np.random.randint(
                        low=low_value[i],
                        high=high_value[i]),
                    self.dim)
            all_feats = np.array(list(range(self.dim)), dtype=np.int32)

            # Use boolean indexing to avoid deprecated nonzero on 0d arrays
            mask = (self.minimum != self.maximum)
            
            # Handle the case when mask is a scalar boolean (occurs when minimum/maximum are scalars)
            if isinstance(mask, bool):
                # If mask is True, use all features
                if mask:
                    valid_indices = np.arange(self.dim)
                else:
                    valid_indices = np.array([], dtype=int)
            else:
                # If mask is an array, proceed as normal
                valid_indices = np.where(mask)[0]
            
            # Select from all_feats using these indices
            choice_feats = all_feats[valid_indices]
            
            # Ensure we have enough features to choose from
            if len(choice_feats) > 0:
                sel_V = np.random.choice(
                    choice_feats, size=min(self.r[i], len(choice_feats)), replace=False)
            else:
                # Fallback if no features meet criteria, select from all features
                sel_V = np.random.choice(
                    all_feats, size=min(self.r[i], len(all_feats)), replace=False)
                
            self.V.append(sel_V)



"""
``pysad`` is a machine learning based anomaly detection framework for streaming data.
"""
# Based on NiLearn package
# License: simplified BSD

# PEP0440 compatible formatted version, see:
# https://www.python.org/dev/peps/pep-0440/
#
# Generic release markers:
# X.Y
# X.Y.Z # For bugfix releases
#
# Admissible pre-release markers:
# X.YaN # Alpha release
# X.YbN # Beta release
# X.YrcN # Release Candidate
# X.Y # Final release
#
# Dev branch marker is: 'X.Y.dev' or 'X.Y.devN' where N is an integer.
# 'X.Y.dev0' is the canonical version of 'X.Y.dev'
#

__version__ = '0.3.3'



import scipy
import numpy as np
from pysad.core.base_model import BaseModel
from pysad.utils.window import Window


class ExactStorm(BaseModel):
    """The Exact-STORM method :cite:`angiulli2007detecting`. This method assigns anomaly score that is the mean of distances to the instances in window of length `window_size` with distnaces less than `max_radius`. Note that the decision making in :cite:`angiulli2007detecting` is not implemented.

            Args:
                window_size : int (Default=10000)
                    The number of instances in the window to score.
                max_radius : float (Default=0.1)
                    Maximum radius for the near instance selection.
    """

    def __init__(self, window_size=10000, max_radius=0.1):
        self.max_radius = max_radius
        self.window = Window(window_size=window_size)

    def fit_partial(self, X, y=None):
        """Fits the model to next instance. Simply, adds the instance to the window.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to fit.
            y (int): Ignored since the model is unsupervised (Default=None).

        Returns:
            object: self.
        """
        self.window.update(X)

        return self

    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            float: The anomalousness score of the input instance.
        """
        window = self.window.get()[:-1]
        if len(window) == 0:
            return 0.0

        dists = scipy.spatial.distance.cdist(window, [X])

        return np.mean(dists < self.max_radius)



from pysad.core.base_metric import BaseMetric
from pysad.statistics.average_meter import AverageMeter


class WindowedMetric(BaseMetric):
    """A helper class to evaluate windowed metrics. The distributions of the streaming model scores often change due to model collapse (i.e. becoming closer to the always loss=0) or appearing nonstationarities. Thus, the metrics such as ROC or AUC scores may change drastically. To prevent their effect, this class creates windows of size `window_size`. After each `window_size`th object, a new instance of the `metric_cls` is being created. Lastly, the metrics from all windows are averaged :cite:`xstream,gokcesu2017online`.

    Args:
        metric_cls (class): The metric class to be windowed.
        window_size (int): The window size.
        ignore_nonempty_last (bool): Whether to ignore the score of the nonempty last window. Note that the empty last window is always ignored.
    """

    def __init__(
            self,
            metric_cls,
            window_size,
            ignore_nonempty_last=True,
            **kwargs):
        super().__init__()
        self.ignore_nonempty_last = ignore_nonempty_last
        self.window_size = window_size
        self.metric_cls = metric_cls

        self.metric = self._init_metric(**kwargs)

        self.score_meter = AverageMeter()
        self.step = 0
        self.num_windows = 1

    def _init_metric(self, **kwargs):

        return self.metric_cls(**kwargs)

    def update(self, y_true, y_pred):
        """Updates the score with new true label and predicted score/label.

        Args:
            y_true : float
                The ground truth score for the incoming instance.
            y_pred : float
                The predicted score for the incoming instance.

        Returns:
            object: self.
        """
        self.step += 1
        self.metric.update(y_true, y_pred)

        if self.step % self.window_size == 0:
            self.num_windows += 1
            score = self.metric.get()
            self.score_meter.update(score)
            self.metric = self._init_metric()

        return self

    def get(self):
        """Obtains the averaged score.

        Returns:
            float: The average score of the windows.
        """
        if self.num_windows == 1:
            return self.metric.get()
        elif not self.ignore_nonempty_last and self.step % self.window_size != 0:
            return (self.metric.get() + self.score_meter.get() * (self.num_windows - 1)) / self.num_windows
        else:
            return self.score_meter.get()



from abc import abstractmethod, ABC


class BaseMetric(ABC):
    """Abstract base class for metrics.
    """

    def __init__(self):
        self.score = None

    @abstractmethod
    def update(self, y_true, y_pred):
        """Updates the metric with given true and predicted value for a timestep.

        Args:
            y_true (int): Ground truth class. Either 1 or 0.
            y_pred (float): Predicted class or anomaly score. Higher values correspond to more anomalousness and lower values correspond to more normalness.
        """
        pass

    @abstractmethod
    def get(self):
        """Gets the current value of the score. Note that some methods such as AUPR and AUROC gives exception when used with only one class exist in the list of previous y_trues.

        Returns:
            float: The current score.
        """
        return self.score



"""
Tests for pysad.utils module functions and classes.
"""
import numpy as np


class TestUtilsFunctions:
    """Tests for utility functions in pysad.utils."""
    
    def test_fix_seed(self):
        """Test fix_seed function for reproducible randomness."""
        from pysad.utils import fix_seed
        
        # Test with specific seed
        fix_seed(42)
        random_val1 = np.random.random()
        
        fix_seed(42)
        random_val2 = np.random.random()
        
        # Should be identical with same seed
        assert random_val1 == random_val2
        
        # Test with different seed
        fix_seed(123)
        random_val3 = np.random.random()
        
        # Should be different with different seed
        assert random_val1 != random_val3
        
    def test_get_minmax_array(self):
        """Test get_minmax_array function."""
        from pysad.utils import get_minmax_array
        
        # Test with 2D array
        X = np.array([[1.0, 5.0], [3.0, 2.0], [0.0, 8.0]])
        min_vals, max_vals = get_minmax_array(X)
        
        expected_min = np.array([0.0, 2.0])
        expected_max = np.array([3.0, 8.0])
        
        np.testing.assert_array_equal(min_vals, expected_min)
        np.testing.assert_array_equal(max_vals, expected_max)
        
    def test_get_minmax_array_single_feature(self):
        """Test get_minmax_array with single feature."""
        from pysad.utils import get_minmax_array
        
        X = np.array([[1.0], [3.0], [0.0], [5.0]])
        min_vals, max_vals = get_minmax_array(X)
        
        assert min_vals[0] == 0.0
        assert max_vals[0] == 5.0
        assert len(min_vals) == 1
        assert len(max_vals) == 1
        
    def test_get_minmax_array_identical_values(self):
        """Test get_minmax_array with identical values."""
        from pysad.utils import get_minmax_array
        
        X = np.array([[2.0, 3.0], [2.0, 3.0], [2.0, 3.0]])
        min_vals, max_vals = get_minmax_array(X)
        
        expected_min = np.array([2.0, 3.0])
        expected_max = np.array([2.0, 3.0])
        
        np.testing.assert_array_equal(min_vals, expected_min)
        np.testing.assert_array_equal(max_vals, expected_max)
        
    def test_get_minmax_scalar(self):
        """Test get_minmax_scalar function."""
        from pysad.utils import get_minmax_scalar
        
        # Test with 1D array
        x = np.array([1.0, 3.0, 0.0, 5.0, 2.0])
        min_val, max_val = get_minmax_scalar(x)
        
        assert min_val == 0.0
        assert max_val == 5.0
        
    def test_get_minmax_scalar_2d_array(self):
        """Test get_minmax_scalar with 2D array."""
        from pysad.utils import get_minmax_scalar
        
        x = np.array([[1.0, 3.0], [0.0, 5.0]])
        min_val, max_val = get_minmax_scalar(x)
        
        assert min_val == 0.0
        assert max_val == 5.0
        
    def test_get_minmax_scalar_single_value(self):
        """Test get_minmax_scalar with single value."""
        from pysad.utils import get_minmax_scalar
        
        x = np.array([42.0])
        min_val, max_val = get_minmax_scalar(x)
        
        assert min_val == 42.0
        assert max_val == 42.0
        
    def test_iterate_without_labels(self):
        """Test _iterate function without labels."""
        from pysad.utils import _iterate
        
        X = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        
        results = list(_iterate(X))
        
        assert len(results) == 3
        for i, (xi, yi) in enumerate(results):
            np.testing.assert_array_equal(xi, X[i])
            assert yi is None
            
    def test_iterate_with_labels(self):
        """Test _iterate function with labels."""
        from pysad.utils import _iterate
        
        X = np.array([[1.0, 2.0], [3.0, 4.0]])
        y = np.array([0, 1])
        
        results = list(_iterate(X, y))
        
        assert len(results) == 2
        for i, (xi, yi) in enumerate(results):
            np.testing.assert_array_equal(xi, X[i])
            assert yi == y[i]
            
    def test_iterate_single_instance(self):
        """Test _iterate function with single instance."""
        from pysad.utils import _iterate
        
        X = np.array([[1.0, 2.0]])
        y = np.array([1])
        
        results = list(_iterate(X, y))
        
        assert len(results) == 1
        xi, yi = results[0]
        np.testing.assert_array_equal(xi, X[0])
        assert yi == y[0]
        
    def test_iterate_empty_array(self):
        """Test _iterate function with empty array."""
        from pysad.utils import _iterate
        
        X = np.array([]).reshape(0, 2)
        
        results = list(_iterate(X))
        
        assert len(results) == 0


class TestWindow:
    """Tests for Window class."""
    
    def test_window_initialization(self):
        """Test Window initialization."""
        from pysad.utils import Window
        
        window = Window(window_size=5)
        assert window.window_size == 5
        assert window.window == []
        
    def test_window_update_within_size(self):
        """Test Window update within size limit."""
        from pysad.utils import Window
        
        window = Window(window_size=3)
        
        window.update(1.0)
        window.update(2.0)
        
        assert window.get() == [1.0, 2.0]
        assert len(window.get()) == 2
        
    def test_window_update_exceeds_size(self):
        """Test Window update when exceeding size."""
        from pysad.utils import Window
        
        window = Window(window_size=3)
        
        # Add more items than window size
        for i in range(5):
            window.update(float(i))
            
        result = window.get()
        assert len(result) == 3
        assert result == [2.0, 3.0, 4.0]  # Should keep last 3
        
    def test_window_single_size(self):
        """Test Window with size 1."""
        from pysad.utils import Window
        
        window = Window(window_size=1)
        
        window.update(1.0)
        assert window.get() == [1.0]
        
        window.update(2.0)
        assert window.get() == [2.0]
        
    def test_window_get_copy(self):
        """Test that Window.get() returns the actual list."""
        from pysad.utils import Window
        
        window = Window(window_size=5)
        window.update(1.0)
        window.update(2.0)
        
        result1 = window.get()
        result2 = window.get()
        
        # Should be the same list
        assert result1 is result2
        
    def test_unlimited_window_initialization(self):
        """Test UnlimitedWindow initialization."""
        from pysad.utils.window import UnlimitedWindow
        
        window = UnlimitedWindow()
        assert window.window_size is None
        assert window.window == []
        
    def test_unlimited_window_growth(self):
        """Test UnlimitedWindow unlimited growth."""
        from pysad.utils.window import UnlimitedWindow
        
        window = UnlimitedWindow()
        
        # Add many items
        for i in range(100):
            window.update(float(i))
            
        result = window.get()
        assert len(result) == 100
        assert result[0] == 0.0
        assert result[-1] == 99.0
        
    def test_unlimited_window_no_size_limit(self):
        """Test that UnlimitedWindow doesn't remove old items."""
        from pysad.utils.window import UnlimitedWindow
        
        window = UnlimitedWindow()
        
        for i in range(10):
            window.update(float(i))
            
        # All items should be present
        result = window.get()
        expected = [float(i) for i in range(10)]
        assert result == expected


class TestArrayStreamer:
    """Tests for ArrayStreamer class."""
    
    def test_array_streamer_initialization(self):
        """Test ArrayStreamer initialization."""
        from pysad.utils import ArrayStreamer
        
        streamer = ArrayStreamer(shuffle=False)
        assert hasattr(streamer, 'shuffle')
        assert streamer.shuffle is False
        
        streamer = ArrayStreamer(shuffle=True)
        assert streamer.shuffle is True
        
    def test_array_streamer_iter_without_labels(self):
        """Test ArrayStreamer iteration without labels."""
        from pysad.utils import ArrayStreamer
        
        streamer = ArrayStreamer(shuffle=False)
        X = np.array([[1, 2], [3, 4], [5, 6]])
        
        results = list(streamer.iter(X))
        assert len(results) == 3
        for i, xi in enumerate(results):
            np.testing.assert_array_equal(xi, X[i])
            
    def test_array_streamer_iter_with_labels(self):
        """Test ArrayStreamer iteration with labels."""
        from pysad.utils import ArrayStreamer
        
        streamer = ArrayStreamer(shuffle=False)
        X = np.array([[1, 2], [3, 4]])
        y = np.array([0, 1])
        
        results = list(streamer.iter(X, y))
        assert len(results) == 2
        for i, (xi, yi) in enumerate(results):
            np.testing.assert_array_equal(xi, X[i])
            assert yi == y[i]
            
    def test_array_streamer_shuffle(self):
        """Test ArrayStreamer with shuffle."""
        from pysad.utils import ArrayStreamer
        
        # Set seed for reproducible shuffle
        np.random.seed(42)
        streamer = ArrayStreamer(shuffle=True)
        X = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])
        
        results = list(streamer.iter(X))
        assert len(results) == 4
        
        # Results should contain all original data but potentially in different order
        result_set = {tuple(xi) for xi in results}
        original_set = {tuple(X[i]) for i in range(len(X))}
        assert result_set == original_set


class TestPandasStreamer:
    """Tests for PandasStreamer class."""
    
    def test_pandas_streamer_initialization(self):
        """Test PandasStreamer initialization."""
        from pysad.utils import PandasStreamer
        
        streamer = PandasStreamer(shuffle=False)
        assert hasattr(streamer, 'shuffle')
        assert streamer.shuffle is False
        
    def test_pandas_streamer_with_dataframe(self):
        """Test PandasStreamer with pandas DataFrame."""
        import pandas as pd
        from pysad.utils import PandasStreamer
        
        # Create test DataFrame
        df = pd.DataFrame({
            'feature1': [1.0, 2.0, 3.0],
            'feature2': [4.0, 5.0, 6.0]
        })
        
        streamer = PandasStreamer(shuffle=False)
        results = list(streamer.iter(df))
        
        assert len(results) == 3
        for i, row in enumerate(results):
            assert len(row) == 2
            assert row[0] == df.iloc[i, 0]
            assert row[1] == df.iloc[i, 1]
            
    def test_pandas_streamer_with_labels(self):
        """Test PandasStreamer with labels."""
        import pandas as pd
        from pysad.utils import PandasStreamer
        
        # Create test data
        df = pd.DataFrame({
            'feature1': [1.0, 2.0],
            'feature2': [3.0, 4.0]
        })
        labels = pd.Series([0, 1])
        
        streamer = PandasStreamer(shuffle=False)
        results = list(streamer.iter(df, labels))
        
        assert len(results) == 2
        for i, (row, label) in enumerate(results):
            assert len(row) == 2
            assert label == labels.iloc[i]



import numpy as np
from scipy.cluster.hierarchy import linkage, fcluster, to_tree

# A helper class for KitNET which performs a correlation-based incremental clustering of the dimensions in X
# n: the number of dimensions in the dataset
# For more information and citation, please see our NDSS'18 paper: Kitsune: An Ensemble of Autoencoders for Online Network Intrusion Detection
class corClust:
    def __init__(self,n):
        #parameter:
        self.n = n
        #varaibles
        self.c = np.zeros(n) #linear num of features
        self.c_r = np.zeros(n) #linear sum of feature residules
        self.c_rs = np.zeros(n) #linear sum of feature residules
        self.C = np.zeros((n,n)) #partial correlation matrix
        self.N = 0 #number of updates performed

    # x: a numpy vector of length n
    def update(self,x):
        self.N += 1
        self.c += x
        c_rt = x - self.c/self.N
        self.c_r += c_rt
        self.c_rs += c_rt**2
        self.C += np.outer(c_rt,c_rt)

    # creates the current correlation distance matrix between the features
    def corrDist(self):
        c_rs_sqrt = np.sqrt(self.c_rs)
        C_rs_sqrt = np.outer(c_rs_sqrt,c_rs_sqrt)
        C_rs_sqrt[C_rs_sqrt==0] = 1e-100 #this protects against dive by zero erros (occurs when a feature is a constant)
        D = 1-self.C/C_rs_sqrt #the correlation distance matrix
        D[D<0] = 0 #small negatives may appear due to the incremental fashion in which we update the mean. Therefore, we 'fix' them
        return D

    # clusters the features together, having no more than maxClust features per cluster
    def cluster(self,maxClust):
        D = self.corrDist()
        Z = linkage(D[np.triu_indices(self.n, 1)])  # create a linkage matrix based on the distance matrix
        if maxClust < 1:
            maxClust = 1
        if maxClust > self.n:
            maxClust = self.n
        map = self.__breakClust__(to_tree(Z),maxClust)
        return map

    # a recursive helper function which breaks down the dendrogram branches until all clusters have no more than maxClust elements
    def __breakClust__(self,dendro,maxClust):
        if dendro.count <= maxClust: #base case: we found a minimal cluster, so mark it
            return [dendro.pre_order()] #return the origional ids of the features in this cluster
        return self.__breakClust__(dendro.get_left(),maxClust) + self.__breakClust__(dendro.get_right(),maxClust)

# Copyright (c) 2017 Yisroel Mirsky
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


def test_unsupervised_models():
    from pysad.models import RobustRandomCutForest
    from pysad.models import ExactStorm
    from pysad.models import HalfSpaceTrees
    from pysad.models import IForestASD
    from pysad.models import KitNet
    from pysad.models import KNNCAD
    from pysad.models import LODA
    from pysad.models import LocalOutlierProbability
    from pysad.models import MedianAbsoluteDeviation
    from pysad.models import NullModel
    from pysad.models import RandomModel
    from pysad.models import RelativeEntropy
    from pysad.models import RSHash
    from pysad.models import StandardAbsoluteDeviation
    from pysad.models import xStream
    from pysad.models import Inqmad
    import numpy as np
    from pysad.utils import fix_seed
    fix_seed(61)

    X = np.random.rand(150, 1)

    model_classes = {
        RelativeEntropy: {"min_val": 0.0, "max_val": 1.0},
        ExactStorm: {},
        HalfSpaceTrees: {"feature_mins": [0.0], "feature_maxes": [1.0]},
        IForestASD: {},
        KitNet: {},
        KNNCAD: {"probationary_period": 50},
        LODA: {},
        LocalOutlierProbability: { "initial_X": True },
        MedianAbsoluteDeviation: [{}, { "absolute": False }],
        NullModel: {},
        RandomModel: {},
        RSHash: {"feature_mins": [0.0], "feature_maxes": [1.0]},
        StandardAbsoluteDeviation: [{}, {"absolute": False}],
        xStream: {},
        RobustRandomCutForest: {},
        Inqmad: {"input_shape": X.shape[1], "dim_x": 128, "gamma": 100}
    }

    for model_cls, params_dict in model_classes.items():
        if type(params_dict) is dict:
            helper_test_model(X, model_cls, params_dict)
        elif type(params_dict) is list:
            for params in params_dict:
                helper_test_model(X, model_cls, params)


def helper_test_model(X, model_cls, params_dict):

    if "initial_X" in params_dict and params_dict["initial_X"]:
        params_dict["initial_X"] = X[:25, :]
        model = model_cls(params_dict["initial_X"])
        train_X = X[25:, :]
    else:
        train_X = X
        model = model_cls(**params_dict)

    y_pred = model.fit_score(train_X)
    assert y_pred.shape == (train_X.shape[0],)


def test_fit_and_score_separately():
    from pysad.models import xStream
    import numpy as np
    from pysad.utils import fix_seed
    fix_seed(61)

    X = np.random.rand(150, 1)

    model = xStream()

    model = model.fit(X)
    y_pred = model.score(X)
    assert y_pred.shape == (X.shape[0],)


def test_relative_entropy_no_warnings():
    """Test that RelativeEntropy doesn't produce DeprecationWarnings about scalar conversion."""
    import warnings
    import numpy as np
    from pysad.models import RelativeEntropy
    from pysad.utils import fix_seed
    
    fix_seed(61)
    X = np.random.rand(150, 1)
    
    # Capture any DeprecationWarnings about scalar conversion
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        
        model = RelativeEntropy(min_val=0.0, max_val=1.0)
        y_pred = model.fit_score(X)
        
        # Filter for the specific warnings we're concerned about
        scalar_warnings = [warning for warning in w 
                          if issubclass(warning.category, DeprecationWarning) and
                          ('scalar divide' in str(warning.message) or 
                           'Conversion of an array with ndim > 0 to a scalar' in str(warning.message))]
        
        # Assert no scalar conversion warnings occurred
        assert len(scalar_warnings) == 0, f"Found {len(scalar_warnings)} scalar conversion warnings"
        assert y_pred.shape == (X.shape[0],)



from pysad.core.base_statistic import UnivariateStatistic


class AbsStatistic(UnivariateStatistic):
    """The absolute value of the statistic that is tracked.

    Args:
        statistic_cls (class): The class of the statistic to be instiantiated.
        **kwargs (Keyword arguments): The keyword arguments that is input to the statistic_cls.
    """

    def __init__(self, statistic_cls, **kwargs):
        self.statistic_cls = statistic_cls

        self.statistic = self.statistic_cls(**kwargs)

    def update(self, num):
        """Updates the statistic with the value for a timestep.

        Args:
            num (float):  The incoming value, for which the statistic is used.

        Returns:
            object: self.

        """
        self.statistic.update(num)

        return self

    def remove(self, num):
        """Updates the statistic by removing particular value. This method

        Args:
            num (float):  The value to be removed.

        Returns:
            object: self.

        """
        self.statistic.remove(num)

        return self

    def get(self):
        """Method to obtain the tracked statistic.

        Returns:
            float: The statistic.
        """
        return abs(self.statistic.get())



from pysad.core.base_statistic import UnivariateStatistic


class CountMeter(UnivariateStatistic):
    """A simple counter statistic.

    Attrs:
        count (int): The number of items that are used to update the statistic.
    """

    def __init__(self):
        self.count = 0

    def update(self, num):
        """Updates the statistic with the value for a timestep.

        Args:
            num (float): The incoming value, for which the statistic is used.

        Returns:
            object: self.
        """
        self.count += 1

        return self

    def remove(self, num=None):
        """Updates the statistic by removing particular value. This method

        Args:
            num (float): The value to be removed.

        Returns:
            object: self.
        """
        self.count -= 1

        return self

    def get(self):
        """ Method to obtain the tracked statistic.

        Returns:
            float: The statistic.
        """
        return self.count



import math
from heapq import heappush
from pysad.core.base_statistic import UnivariateStatistic
import numpy as np


class MinMeter(UnivariateStatistic):
    """The statistic that keeps track of the minimum value.

        Attrs:
            min (float): The minimum value.
            lst (list[float]): The list of values that are used to update the statistic. It is necessary for windowing operations.
    """

    def __init__(self):
        self.min = math.inf

        self.lst = []

    def update(self, num):
        """Updates the statistic with the value for a timestep.

        Args:
            num (float): The incoming value, for which the statistic is used.

        Returns:
            object: self.
        """
        if num < self.min:
            self.min = num

        heappush(self.lst, num)

        return self

    def remove(self, num):
        """Updates the statistic by removing particular value. This method

        Args:
            num (float): The value to be removed.

        Returns:
            object: self.
        """
        self.lst.remove(num)

        if len(self.lst) > 0:
            self.min = np.min(self.lst)
        else:
            self.min = math.inf

        return self

    def get(self):
        """ Method to obtain the tracked statistic.

        Returns:
            float: The statistic.
        """
        return self.min



from pysad.core.base_model import BaseModel
import numpy as np


class LODA(BaseModel):
    """The LODA model :cite:`pevny2016loda` The implemnetation is adapted to the steraming framework from the `PyOD framework <https://pyod.readthedocs.io/en/latest/_modules/pyod/models/loda.html#LODA>`_.

        Args:
            num_bins (int): The number of bins of the histogram.
            num_random_cuts (int): The number of random cuts.
    """

    def __init__(self, num_bins=10, num_random_cuts=100):
        self.to_init = True
        self.n_bins = num_bins
        self.n_random_cuts = num_random_cuts

    def fit_partial(self, X, y=None):
        """Fits the model to next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to fit.
            y (int): Ignored since the model is unsupervised (Default=None).

        Returns:
            object: Returns the self.
        """
        if self.to_init:
            self.num_features = X.shape[0]
            self.weights = np.ones(
                self.n_random_cuts,
                dtype=np.float64) / self.n_random_cuts
            self.projections_ = np.random.randn(
                self.n_random_cuts, self.num_features)
            self.histograms_ = np.zeros((self.n_random_cuts, self.n_bins))
            self.limits_ = np.zeros((self.n_random_cuts, self.n_bins + 1))

            n_nonzero_components = np.sqrt(self.num_features)
            self.n_zero_components = self.num_features - \
                np.int32(n_nonzero_components)

            self.to_init = False

        X = X.reshape(1, -1)

        for i in range(self.n_random_cuts):
            rands = np.random.permutation(self.num_features)[
                :self.n_zero_components]
            self.projections_[i, rands] = 0.
            projected_data = self.projections_[i, :].dot(X.T)
            self.histograms_[i, :], self.limits_[i, :] = np.histogram(
                projected_data, bins=self.n_bins, density=False)
            self.histograms_[i, :] += 1e-12
            self.histograms_[i, :] /= np.sum(self.histograms_[i, :])

        return self

    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            float: The anomalousness score of the input instance.
        """
        X = X.reshape(1, -1)

        pred_scores = np.zeros([X.shape[0], 1])
        for i in range(self.n_random_cuts):
            projected_data = self.projections_[i, :].dot(X.T)
            inds = np.searchsorted(self.limits_[i, :self.n_bins - 1],
                                   projected_data, side='left')
            pred_scores[:, 0] += -self.weights[i] * np.log(
                self.histograms_[i, inds])
        pred_scores /= self.n_random_cuts

        return pred_scores.ravel()



"""
Tests for pysad.core base classes and abstract methods.
"""
import pytest
import numpy as np

from pysad.core.base_model import BaseModel
from pysad.core.base_metric import BaseMetric
from pysad.core.base_postprocessor import BasePostprocessor
from pysad.core.base_statistic import BaseStatistic
from pysad.core.base_streamer import BaseStreamer
from pysad.core.base_transformer import BaseTransformer


class ConcreteModel(BaseModel):
    """Test implementation of BaseModel for testing purposes."""
    
    def __init__(self):
        self.fitted_data = []
        self.score_count = 0
        
    def fit_partial(self, X, y=None):
        self.fitted_data.append((X.copy(), y))
        return self
        
    def score_partial(self, X):
        self.score_count += 1
        # Return a simple score based on L2 norm
        return np.linalg.norm(X)


class ConcreteMetric(BaseMetric):
    """Test implementation of BaseMetric for testing purposes."""
    
    def __init__(self):
        self.values = []
        
    def update(self, y_true, y_pred):
        self.values.append((y_true, y_pred))
        
    def get(self):
        if not self.values:
            return 0.0
        # Return accuracy for testing
        correct = sum(1 for y_true, y_pred in self.values if y_true == (y_pred > 0.5))
        return correct / len(self.values)


class ConcretePostprocessor(BasePostprocessor):
    """Test implementation of BasePostprocessor for testing purposes."""
    
    def __init__(self, multiplier=2.0):
        self.multiplier = multiplier
        self.fitted_scores = []
        
    def fit_partial(self, score):
        self.fitted_scores.append(score)
        return self
        
    def transform_partial(self, score=None):
        return score * self.multiplier if score is not None else 0.0


class ConcreteStatistic(BaseStatistic):
    """Test implementation of BaseStatistic for testing purposes."""
    
    def __init__(self):
        self.values = []
        
    def update(self, num):
        self.values.append(num)
        
    def get(self):
        return np.mean(self.values) if self.values else 0.0


class ConcreteStreamer(BaseStreamer):
    """Test implementation of BaseStreamer for testing purposes."""
    
    def __init__(self, shuffle=False):
        self.shuffle = shuffle
        
    def iter(self, X, y=None):
        indices = np.arange(len(X))
        if self.shuffle:
            np.random.shuffle(indices)
            
        if y is None:
            for i in indices:
                yield X[i]
        else:
            for i in indices:
                yield X[i], y[i]


class ConcreteTransformer(BaseTransformer):
    """Test implementation of BaseTransformer for testing purposes."""
    
    def __init__(self):
        self.fitted_data = []
        
    def fit_partial(self, X):
        self.fitted_data.append(X.copy())
        return self
        
    def transform_partial(self, X):
        # Simple transformation: add 1.0
        return X + 1.0


class TestBaseModel:
    """Tests for BaseModel abstract class and its concrete methods."""
    
    def test_abstract_methods_exist(self):
        """Test that BaseModel has the required abstract methods."""
        model = ConcreteModel()
        assert hasattr(model, 'fit_partial')
        assert hasattr(model, 'score_partial')
        assert callable(model.fit_partial)
        assert callable(model.score_partial)
        
    def test_fit_score_partial(self):
        """Test fit_score_partial method."""
        model = ConcreteModel()
        X = np.array([1.0, 2.0, 3.0])
        y = 1
        
        score = model.fit_score_partial(X, y)
        
        assert len(model.fitted_data) == 1
        assert model.score_count == 1
        np.testing.assert_array_equal(model.fitted_data[0][0], X)
        assert model.fitted_data[0][1] == y
        assert isinstance(score, float)
        
    def test_fit_method(self):
        """Test batch fitting with fit method."""
        model = ConcreteModel()
        X = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        y = np.array([0, 1, 0])
        
        fitted_model = model.fit(X, y)
        
        assert fitted_model is model  # Should return self
        assert len(model.fitted_data) == 3
        for i in range(3):
            np.testing.assert_array_equal(model.fitted_data[i][0], X[i])
            assert model.fitted_data[i][1] == y[i]
            
    def test_fit_method_without_labels(self):
        """Test batch fitting without labels."""
        model = ConcreteModel()
        X = np.array([[1.0, 2.0], [3.0, 4.0]])
        
        fitted_model = model.fit(X)
        
        assert fitted_model is model
        assert len(model.fitted_data) == 2
        for i in range(2):
            np.testing.assert_array_equal(model.fitted_data[i][0], X[i])
            assert model.fitted_data[i][1] is None
            
    def test_score_method(self):
        """Test batch scoring with score method."""
        model = ConcreteModel()
        X = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
        
        scores = model.score(X)
        
        assert len(scores) == 3
        assert model.score_count == 3
        assert isinstance(scores, np.ndarray)
        assert scores.dtype == np.float64
        
    def test_fit_score_method(self):
        """Test batch fit and score with fit_score method."""
        model = ConcreteModel()
        X = np.array([[1.0, 2.0], [3.0, 4.0]])
        y = np.array([0, 1])
        
        scores = model.fit_score(X, y)
        
        assert len(scores) == 2
        assert len(model.fitted_data) == 2
        assert model.score_count == 2
        assert isinstance(scores, np.ndarray)
        assert scores.dtype == np.float64
        
    def test_fit_score_method_without_labels(self):
        """Test batch fit and score without labels."""
        model = ConcreteModel()
        X = np.array([[1.0, 2.0], [3.0, 4.0]])
        
        scores = model.fit_score(X)
        
        assert len(scores) == 2
        assert len(model.fitted_data) == 2
        for fitted_data in model.fitted_data:
            assert fitted_data[1] is None


class TestBaseMetric:
    """Tests for BaseMetric abstract class."""
    
    def test_abstract_methods_exist(self):
        """Test that BaseMetric has the required abstract methods."""
        metric = ConcreteMetric()
        assert hasattr(metric, 'update')
        assert hasattr(metric, 'get')
        assert callable(metric.update)
        assert callable(metric.get)
        
    def test_update_and_get(self):
        """Test basic update and get functionality."""
        metric = ConcreteMetric()
        
        # Initially should return default value
        assert metric.get() == 0.0
        
        # Add some values
        metric.update(1, 0.8)  # Correct
        metric.update(0, 0.2)  # Correct
        metric.update(1, 0.3)  # Incorrect
        
        accuracy = metric.get()
        assert 0.0 <= accuracy <= 1.0


class TestBasePostprocessor:
    """Tests for BasePostprocessor abstract class."""
    
    def test_abstract_methods_exist(self):
        """Test that BasePostprocessor has the required abstract methods."""
        processor = ConcretePostprocessor()
        assert hasattr(processor, 'fit_partial')
        assert hasattr(processor, 'transform_partial')
        assert callable(processor.fit_partial)
        assert callable(processor.transform_partial)
        
    def test_fit_transform_workflow(self):
        """Test the fit and transform workflow."""
        processor = ConcretePostprocessor(multiplier=3.0)
        
        # Fit on some scores
        processor.fit_partial(1.0)
        processor.fit_partial(2.0)
        
        assert len(processor.fitted_scores) == 2
        
        # Transform a score
        result = processor.transform_partial(5.0)
        assert result == 15.0  # 5.0 * 3.0
        
    def test_transform_without_score(self):
        """Test transform_partial with None score."""
        processor = ConcretePostprocessor()
        result = processor.transform_partial()
        assert result == 0.0


class TestBaseStatistic:
    """Tests for BaseStatistic abstract class."""
    
    def test_abstract_methods_exist(self):
        """Test that BaseStatistic has the required abstract methods."""
        stat = ConcreteStatistic()
        assert hasattr(stat, 'update')
        assert hasattr(stat, 'get')
        assert callable(stat.update)
        assert callable(stat.get)
        
    def test_update_and_get(self):
        """Test basic update and get functionality."""
        stat = ConcreteStatistic()
        
        # Initially should return default value
        assert stat.get() == 0.0
        
        # Add some values
        stat.update(1.0)
        stat.update(2.0)
        stat.update(3.0)
        
        mean_value = stat.get()
        assert mean_value == 2.0


class TestBaseStreamer:
    """Tests for BaseStreamer abstract class."""
    
    def test_abstract_methods_exist(self):
        """Test that BaseStreamer has the required abstract methods."""
        streamer = ConcreteStreamer()
        assert hasattr(streamer, 'iter')
        assert callable(streamer.iter)
        
    def test_iter_without_labels(self):
        """Test iteration without labels."""
        streamer = ConcreteStreamer(shuffle=False)
        X = np.array([[1, 2], [3, 4], [5, 6]])
        
        results = list(streamer.iter(X))
        assert len(results) == 3
        for i, x in enumerate(results):
            np.testing.assert_array_equal(x, X[i])
            
    def test_iter_with_labels(self):
        """Test iteration with labels."""
        streamer = ConcreteStreamer(shuffle=False)
        X = np.array([[1, 2], [3, 4]])
        y = np.array([0, 1])
        
        results = list(streamer.iter(X, y))
        assert len(results) == 2
        for i, (x, y_val) in enumerate(results):
            np.testing.assert_array_equal(x, X[i])
            assert y_val == y[i]
            
    def test_iter_with_shuffle(self):
        """Test iteration with shuffle enabled."""
        np.random.seed(42)
        streamer = ConcreteStreamer(shuffle=True)
        X = np.array([[1, 2], [3, 4], [5, 6]])
        
        results = list(streamer.iter(X))
        assert len(results) == 3
        # With shuffle, order should potentially be different


class TestBaseTransformer:
    """Tests for BaseTransformer abstract class."""
    
    def test_abstract_methods_exist(self):
        """Test that BaseTransformer has the required abstract methods."""
        transformer = ConcreteTransformer()
        assert hasattr(transformer, 'fit_partial')
        assert hasattr(transformer, 'transform_partial')
        assert callable(transformer.fit_partial)
        assert callable(transformer.transform_partial)
        
    def test_fit_transform_workflow(self):
        """Test the fit and transform workflow."""
        transformer = ConcreteTransformer()
        X = np.array([1.0, 2.0, 3.0])
        
        # Fit
        fitted_transformer = transformer.fit_partial(X)
        assert fitted_transformer is transformer
        assert len(transformer.fitted_data) == 1
        np.testing.assert_array_equal(transformer.fitted_data[0], X)
        
        # Transform
        result = transformer.transform_partial(X)
        expected = X + 1.0
        np.testing.assert_array_equal(result, expected)


class TestAbstractClassInstantiation:
    """Test that abstract base classes cannot be instantiated directly."""
    
    def test_base_model_abstract(self):
        """Test that BaseModel cannot be instantiated directly."""
        with pytest.raises(TypeError):
            BaseModel()
            
    def test_base_metric_abstract(self):
        """Test that BaseMetric cannot be instantiated directly."""
        with pytest.raises(TypeError):
            BaseMetric()
            
    def test_base_postprocessor_abstract(self):
        """Test that BasePostprocessor cannot be instantiated directly."""
        with pytest.raises(TypeError):
            BasePostprocessor()
            
    def test_base_statistic_abstract(self):
        """Test that BaseStatistic can be instantiated but has no abstract methods."""
        # BaseStatistic is abstract but has no abstract methods, so it can be instantiated
        stat = BaseStatistic()
        assert isinstance(stat, BaseStatistic)
            
    def test_base_streamer_abstract(self):
        """Test that BaseStreamer cannot be instantiated directly."""
        with pytest.raises(TypeError):
            BaseStreamer()
            
    def test_base_transformer_abstract(self):
        """Test that BaseTransformer cannot be instantiated directly."""
        with pytest.raises(TypeError):
            BaseTransformer()



from pysad.core.base_transformer import BaseTransformer
import numpy as np


class InstanceUnitNormScaler(BaseTransformer):
    """A scaler that makes the instance feature vector's norm equal to 1, i.e., the unit vector.

    Args:
        pow (float): The power, for which the norm is calculated. pow=2 is equivalent to the euclidean distance.
    """

    def __init__(self, pow=2):
        super().__init__(-1)
        self.pow = pow

    def fit_partial(self, X):
        """Fits particular (next) timestep's features to train the scaler.

        Args:
            X (np.float64 array of shape (num_features,)): Input feature vector.

        Returns:
            object: self.
        """
        return self

    def transform_partial(self, X):
        """Scales particular (next) timestep's vector.

        Args:
            X (np.float64 array of shape (num_features,)): Input feature vector.

        Returns:
            scaled_X (np.float64 array of shape (features,)): Scaled feature vector.
        """
        X_norm = np.linalg.norm(X, ord=self.pow)

        return X / X_norm



"""
The :mod:`pysad.models` module includes anomaly detection models to score the anomalousness of instances.
"""
from .half_space_trees import HalfSpaceTrees
from .iforest_asd import IForestASD
from .kitnet import KitNet
from .knn_cad import KNNCAD
from .loda import LODA
from .loop import LocalOutlierProbability
from .median_absolute_deviation import MedianAbsoluteDeviation
from .null_model import NullModel
from .perfect_model import PerfectModel
from .random_model import RandomModel
from .relative_entropy import RelativeEntropy
from .robust_random_cut_forest import RobustRandomCutForest
from .rs_hash import RSHash
from .standard_absolute_deviation import StandardAbsoluteDeviation
from .xstream import xStream
from .exact_storm import ExactStorm

# Try to import Inqmad - this requires JAX dependencies which are optional
try:
    from .inqmad import Inqmad
    _has_inqmad = True
except (ImportError, NameError, AttributeError):
    # Handle missing dependencies, undefined names, and JAX-NumPy compatibility issues
    _has_inqmad = False

__all__ = ["ExactStorm", "HalfSpaceTrees", "IForestASD", "KitNet", "KNNCAD", "LODA", "LocalOutlierProbability", "MedianAbsoluteDeviation", "NullModel", "PerfectModel", "RandomModel", "RelativeEntropy", "RobustRandomCutForest", "RSHash", "StandardAbsoluteDeviation", "xStream"]

# Add Inqmad to __all__ if available
if _has_inqmad:
    __all__.append("Inqmad")




def test_standard_absolute_deviation():
    from pysad.models import StandardAbsoluteDeviation
    import numpy as np
    from numpy.testing import assert_raises
    from pysad.utils import fix_seed

    fix_seed(61)
    X = np.random.rand(150, 1)

    model = StandardAbsoluteDeviation(substracted_statistic="mean")
    model = model.fit(X)
    y_pred = model.score(X)
    assert y_pred.shape == (X.shape[0],)

    model = StandardAbsoluteDeviation(substracted_statistic="median")
    model = model.fit(X)
    y_pred = model.score(X)
    assert y_pred.shape == (X.shape[0],)

    with assert_raises(ValueError):
        StandardAbsoluteDeviation(substracted_statistic="asd")

    with assert_raises(ValueError):
        StandardAbsoluteDeviation(substracted_statistic=None)



def test_perfect_model():
    import numpy as np
    from pysad.models import PerfectModel
    from pysad.utils import fix_seed
    fix_seed(61)

    model = PerfectModel()
    y1 = np.random.randint(0, 2, 100)
    y = np.random.randint(0, 2, 100)
    X = np.random.rand(100)
    y_pred = model.fit_score(X, y)
    print(y_pred)
    assert np.all(np.isclose(y, y_pred))
    assert not np.all(np.isclose(y1, y_pred))



from abc import ABCMeta, abstractmethod
from sklearn.metrics import recall_score, precision_score, roc_auc_score, average_precision_score
from pysad.core.base_metric import BaseMetric


class BaseSKLearnMetric(BaseMetric, metaclass=ABCMeta):
    """Abstract base class to wrap the sklearn metrics.
    """

    def __init__(self):
        self.y_true = []
        self.y_pred = []

    def update(self, y_true, y_pred):
        """Updates the metric with given true and predicted value for a timestep.

        Args:
            y_true (int): Ground truth class. Either 1 or 0.
            y_pred (float): Predicted class or anomaly score. Higher values correspond to more anomalousness and lower values correspond to more normalness.
        """
        self.y_true.append(y_true)
        self.y_pred.append(y_pred)

    def get(self):
        """Gets the current value of the score.

        Returns:
            float: The current score.
        """
        score = self._evaluate(self.y_true, self.y_pred)

        return score

    @abstractmethod
    def _evaluate(self, y_true, y_pred):
        """Abstract method to be filled with the sklearn metric.

        Args:
            y_true (list[int]): Ground truth classes.
            y_pred (list[float]): Predicted classes or scores.
        """
        pass


class PrecisionMetric(BaseSKLearnMetric):
    """Precision wrapper class for sklearn.
    """

    def _evaluate(self, y_true, y_pred):
        return precision_score(y_true, y_pred)


class RecallMetric(BaseSKLearnMetric):
    """Recall wrapper class for sklearn.
    """

    def _evaluate(self, y_true, y_pred):
        return recall_score(y_true, y_pred)


class AUROCMetric(BaseSKLearnMetric):
    """Area under roc curve wrapper class for sklearn.
    """

    def _evaluate(self, y_true, y_pred):
        return roc_auc_score(y_true, y_pred)


class AUPRMetric(BaseSKLearnMetric):
    """Area under PR curve wrapper class for sklearn.
    """

    def _evaluate(self, y_true, y_pred):
        return average_precision_score(y_true, y_pred)



from heapq import heappush
from pysad.core.base_statistic import UnivariateStatistic


class MedianMeter(UnivariateStatistic):
    """The statistic that keeps track of the median.

        Attrs:
            num_items (int): The number of items that are used to update the statistic.
            lst (list[float]): The list of values that are used to update the statistic. It is necessary for windowing operations.
    """

    def __init__(self):
        self.lst = []
        self.num_items = 0

    def update(self, num):
        """Updates the statistic with the value for a timestep.

        Args:
            num (float): The incoming value, for which the statistic is used.

        Returns:
            object: self.
        """
        heappush(self.lst, num)
        self.num_items += 1

        return self

    def remove(self, num):
        """Updates the statistic by removing particular value. This method

        Args:
            num (float): The value to be removed.

        Returns:
            object: self.
        """
        self.lst.remove(num)
        self.num_items -= 1

        return self

    def get(self):
        """ Method to obtain the tracked statistic.

        Returns:
            float: The statistic.
        """
        self.lst = sorted(self.lst)
        if self.num_items % 2 == 0:
            return (self.lst[self.num_items // 2] + self.lst[self.num_items // 2 - 1]) / 2
        else:
            return self.lst[self.num_items // 2]



import numpy as np
from pysad.core.base_transformer import BaseTransformer


class StreamhashProjector(BaseTransformer):
    """Streamhash projection method  from Manzoor et. al.that is similar (or equivalent) to SparseRandomProjection. :cite:`xstream` The implementation is taken from the `cmuxstream-core repository <https://github.com/cmuxstream/cmuxstream-core>`_.

        Args:
            num_components (int): The number of dimensions that the target will be projected into.
            density (float): Density parameter of the streamhash projection.
    """

    def __init__(self, num_components, density=1 / 3.0):
        super().__init__(num_components)
        self.keys = np.arange(0, num_components, 1)
        self.constant = np.sqrt(1. / density) / np.sqrt(num_components)
        self.density = density
        self.n_components = num_components

    def fit_partial(self, X):
        """Fits particular (next) timestep's features to train the projector.

        Args:
            X (np.float64 array of shape (n_components,)): Input feature vector.

        Returns:
            object: self.
        """
        return self

    def transform_partial(self, X):
        """Projects particular (next) timestep's vector to (possibly) lower dimensional space.

        Args:
            X (np.float64 array of shape (num_features,)): Input feature vector.

        Returns:
            projected_X (np.float64 array of shape (num_components,)): Projected feature vector.
        """
        X = X.reshape(1, -1)

        ndim = X.shape[1]

        feature_names = [str(i) for i in range(ndim)]

        R = np.array([[self._hash_string(k, f)
                       for f in feature_names]
                      for k in self.keys])

        Y = np.dot(X, R.T).squeeze()

        return Y

    def _hash_string(self, k, s):
        import mmh3
        hash_value = int(mmh3.hash(s, signed=False, seed=k)) / (2.0 ** 32 - 1)
        s = self.density
        if hash_value <= s / 2.0:
            return -1 * self.constant
        elif hash_value <= s:
            return self.constant
        else:
            return 0



import numpy as np
from pysad.core.base_model import BaseModel
from pysad.transform.projection.streamhash_projector import StreamhashProjector
from pysad.utils import get_minmax_array

# Try to import JAX dependencies, otherwise define a flag to indicate they're missing
try:
    import warnings
    # Suppress numpy.core deprecation warnings
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="numpy.core is deprecated")
        import jax
        from jax import jit
        import jax.numpy as jnp
    JAX_AVAILABLE = True
except (ImportError, AttributeError):
    # Handle both missing JAX and JAX-NumPy compatibility issues
    JAX_AVAILABLE = False

from functools import partial
from sklearn.kernel_approximation import RBFSampler



class Inqmad(BaseModel):
    """The Inqmad model for row-streaming data :cite:`xstream`. (a) an initial normal stream data point is captured (b) those points are mapped to a Hilbert space using adaptive Fourier features (AFF) (c) a memory density matrix $\rho_t$ is initialized using the points from the last step and the $\tau$-threshold value is defined (d) the stream of data points arrives (e) each point is mapped to a Hilbert space using (AFF) (f) a quantum measurement is performed between the streaming point and the memory density matrix $\rho_t$ (g) a $\tau$-threshold value is used to classify normal and anomalous points (h) detect whether the point was classified as normal (i) compute the updated memory density matrix $\rho_{t+1}$ using the normal classified streaming point (j) update the memory density matrix $\rho_t$ with the new matrix $\rho_{t+1}.

    Args:
        input_shape (int): number of features
        dim_x (int): random Fourier features dimension 
        gamma (int): kernel parameter for the random Fourier features 
        random_state (int): initial random state for the random Fourier features
        batch_size (int): training samples processed by iteration

    Note:
        This model requires JAX and JAXlib (version 0.6.1 or higher) to be installed. You can install them using:
        `pip install jax>=0.6.1 jaxlib>=0.6.1`
        or via the optional dependency:
        `pip install pysad[inqmad]`
        
        When using NumPy 2.0 or higher, JAX 0.6.1+ is required for compatibility.
    """

    def __init__(
            self,
            input_shape, 
            dim_x, 
            gamma, 
            random_state=42, 
            batch_size = 300
            ):
        if not JAX_AVAILABLE:
            raise ImportError(
                "JAX dependencies are required to use the Inqmad model. "
                "Please install jax and jaxlib via pip: "
                "`pip install jax jaxlib` or `pip install pysad[inqmad]`"
            )
        self.inqmad = InqMeasurement(input_shape, dim_x, gamma, random_state, batch_size)


    def fit_partial(self, X, y=None):
        """Fits the model to next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to fit.
            y (int): Ignored since the model is unsupervised (Default=None).

        Returns:
            object: Returns the self.
        """

        if(X.ndim == 1):
            X = np.expand_dims(X, axis=0)

        self.inqmad.initial_train(jnp.array(X), 1)
        return self

    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            score (float): The anomalousness score of the input instance.
        """
        if(X.ndim == 1):
            X = np.expand_dims(X, axis=0)

        return self.inqmad.predict(X)



class QFeatureMap_rff():
  """The random Fourier features for Inqmad :cite:`inqmad`.

    Args:
        input_shape (int): number of features
        dim (int): random Fourier features dimension 
        gamma (int): kernel parameter for the random Fourier features 
        random_state (int): initial random state for the random Fourier features
    """

  def __init__(
          self,
          input_dim: int,
          dim: int = 100,
          gamma: float = 1,
          random_state=None,
          **kwargs
  ):
      super().__init__(**kwargs)
      self.input_dim = input_dim
      self.dim = dim
      self.gamma = gamma
      self.random_state = random_state
      self.vmap_compute = jax.jit(jax.vmap(self.compute, in_axes=(0, None, None, None), out_axes=0))

  
  def build(self):
    rbf_sampler = RBFSampler(
            gamma=self.gamma,
            n_components=self.dim,
            random_state=self.random_state)
    x = np.zeros(shape=(1, self.input_dim))
    rbf_sampler.fit(x)

    self.rbf_sampler = rbf_sampler
    self.weights = jnp.array(rbf_sampler.random_weights_)
    self.offset = jnp.array(rbf_sampler.random_offset_)
    self.dim = rbf_sampler.get_params()['n_components']

  def update_rff(self, weights, offset):
    self.weights = jnp.array(weights)
    self.offset = jnp.array(offset)


  def get_dim(self, num_features):
    return self.dim

  @staticmethod
  def compute(X, weights, offset, dim):
    vals = jnp.dot(X, weights) + offset
    #vals = jnp.einsum('i,ik->k', X, weights) + offset
    vals = jnp.cos(vals)
    vals *= jnp.sqrt(2.) / jnp.sqrt(dim)
    return vals
    
  @partial(jit, static_argnums=(0,))
  def __call__(self, X):
    vals = self.vmap_compute(X, self.weights, self.offset, self.dim)
    norms = jnp.linalg.norm(vals, axis=1)
    psi = vals / norms[:, jnp.newaxis]
    return psi   


class InqMeasurement():
  """The Inqmad model for row-streaming data :cite:`xstream`. (a) an initial normal stream data point is captured (b) those points are mapped to a Hilbert space using adaptive Fourier features (AFF) (c) a memory density matrix $\rho_t$ is initialized using the points from the last step and the $\tau$-threshold value is defined (d) the stream of data points arrives (e) each point is mapped to a Hilbert space using (AFF) (f) a quantum measurement is performed between the streaming point and the memory density matrix $\rho_t$ (g) a $\tau$-threshold value is used to classify normal and anomalous points (h) detect whether the point was classified as normal (i) compute the updated memory density matrix $\rho_{t+1}$ using the normal classified streaming point (j) update the memory density matrix $\rho_t$ with the new matrix $\rho_{t+1}.

    Args:
        input_shape (int): number of features
        dim_x (int): random Fourier features dimension 
        gamma (int): kernel parammeter for the random Fourier features 
        random_state (int): initial random state for the random Fourier features
        batch_size (int): training samples processed by iteration
    """



  def __init__(self, input_shape, dim_x, gamma, random_state=42, batch_size = 300):
    self.gamma = gamma
    self.dim_x = dim_x
    self.fm_x = QFeatureMap_rff( input_dim=input_shape, dim = dim_x, gamma = gamma, random_state = random_state)
    self.fm_x.build()
    self.num_samples = 0 
    self.train_pure_batch = jax.jit(jax.vmap(self.train_pure, in_axes=(0)))
    self.collapse_batch = jax.jit(jax.vmap(self.collapse, in_axes=(0, None)))
    self.sum_batch = jax.jit(self.sum)
    self.batch_size = batch_size

  @staticmethod
  def train_pure(inputs):
    oper = jnp.einsum(
        '...i,...j->...ij',
        inputs, jnp.conj(inputs),
        optimize='optimal') # shape (b, nx, nx)
    return oper

  @staticmethod
  def sum(rho_res):
    return jnp.sum(rho_res, axis=0) 


  @staticmethod
  @partial(jit, static_argnums=(1,2,3,4))
  def compute_training_jit(batch, alpha, fm_x, train_pure_batch, sum_batch, rho):
      inputs = fm_x(batch)
      rho_res = train_pure_batch(inputs)
      rho_res = sum_batch(rho_res)
      return jnp.add((alpha)*rho_res, (1-alpha)*rho) if rho is not None else rho_res

  @staticmethod
  def compute_training(values, alpha, perm, i, batch_size, fm_x, train_pure_batch, sum_batch, rho, compute_training_jit):
      batch_idx = perm[i * batch_size: (i + 1)*batch_size]
      batch = values[batch_idx, :]
      return compute_training_jit(batch, alpha, fm_x, train_pure_batch, sum_batch, rho)

  def initial_train(self, values, alpha):
    num_batches = InqMeasurement.obtain_params_batches(values, self.batch_size)
    num_train = values.shape[0]
    perm = jnp.arange(num_train)
    for i in range(num_batches):
      if hasattr(self, "rho_res"):
        self.rho_res = self.compute_training(values, alpha, perm, i,
                            self.batch_size, self.fm_x, 
                            self.train_pure_batch, self.sum_batch, 
                            self.rho_res, self.compute_training_jit)
      else:
        self.rho_res = self.compute_training(values, alpha, perm, i, 
                            self.batch_size, self.fm_x,
                            self.train_pure_batch, self.sum_batch, None, 
                            self.compute_training_jit)
    self.num_samples += values.shape[0]  

    

  @staticmethod
  def collapse(inputs, rho_res):
    rho_h = jnp.matmul(jnp.conj(inputs), rho_res)
    rho_res = jnp.einsum(
        '...i, ...i -> ...',
        rho_h, jnp.conj(rho_h), 
        optimize='optimal') # shape (b,)
    return rho_res

  @staticmethod
  def obtain_params_batches(values, batch_size):
    num_train = values.shape[0]
    num_complete_batches, leftover = divmod(num_train, batch_size)
    num_batches = num_complete_batches + bool(leftover)
    return num_batches

  

  @partial(jit, static_argnums=(0,))
  def predict(self, values):
    num_batches = InqMeasurement.obtain_params_batches(values, self.batch_size)
    results = None
    rho_res = self.rho_res / self.num_samples
    num_train = values.shape[0]
    perm = jnp.arange(num_train)
    for i in range(num_batches):
      batch_idx = perm[i * self.batch_size: (i + 1)*self.batch_size]
      batch = values[batch_idx, :]

      inputs = self.fm_x(batch)
      batch_probs = self.collapse_batch(inputs, rho_res)
      results = jnp.concatenate([results, batch_probs], axis=0) if results is not None else batch_probs
    return results




"""
The :mod:`pysad.transform.postprocessing` module includes postprocessors to transform model scores for streaming learning.
"""

from .postprocessors import AveragePostprocessor, MaxPostprocessor, MedianPostprocessor, ZScorePostprocessor
from .running_postprocessors import RunningAveragePostprocessor, RunningMaxPostprocessor, RunningMedianPostprocessor, RunningZScorePostprocessor

__all__ = ["AveragePostprocessor", "MaxPostprocessor", "MedianPostprocessor", "ZScorePostprocessor", "RunningAveragePostprocessor", "RunningMaxPostprocessor", "RunningMedianPostprocessor", "RunningZScorePostprocessor"]



import os
import pytest


@pytest.fixture
def test_path(): # adapted from https://github.com/scikit-multiflow/scikit-multiflow/blob/master/tests/anomaly_detection/test_half_space_trees.py
    return os.path.dirname(os.path.abspath(__file__))



from pysad.core.base_postprocessor import BasePostprocessor
from pysad.statistics.average_meter import AverageMeter
from pysad.statistics.max_meter import MaxMeter
from pysad.statistics.median_meter import MedianMeter
from pysad.statistics.running_statistic import RunningStatistic
from pysad.statistics.variance_meter import VarianceMeter
import numpy as np


class RunningAveragePostprocessor(BasePostprocessor):
    """A postprocessor that convert a score to the average of of all previous scores in the window.

        Args:
            window_size (int): Length of the window
    """

    def __init__(self, window_size):

        self.meter = RunningStatistic(
            statistic_cls=AverageMeter,
            window_size=window_size)

    def fit_partial(self, score):
        """Fits the windowed postprocessor to the (next) timestep's score.

        Args:
            score (float): Input score.

        Returns:
            object: self.
        """
        self.meter.update(score)

        return self

    def transform_partial(self, score=None):
        """Applies postprocessing to the score using the window.

        Args:
            score (float): The input score.

        Returns:
            float: Transformed score.
        """
        return self.meter.get()


class RunningMaxPostprocessor(BasePostprocessor):
    """A postprocessor that convert a score to the maximum of of all previous scores in the window.
        Args:
            window_size (int): Length of the window
    """

    def __init__(self, window_size):

        self.meter = RunningStatistic(
            statistic_cls=MaxMeter,
            window_size=window_size)

    def fit_partial(self, score):
        """Fits the windowed postprocessor to the (next) timestep's score.

        Args:
            score (float): Input score.

        Returns:
            object: self.
        """
        self.meter.update(score)

        return self

    def transform_partial(self, score=None):
        """Applies postprocessing to the score using the window.

        Args:
            score (float): The input score.

        Returns:
            float: Transformed score.
        """
        return self.meter.get()


class RunningMedianPostprocessor(BasePostprocessor):
    """A postprocessor that convert a score to the median of of all previous scores in the window.
        Args:
            window_size (int): Length of the window
    """

    def __init__(self, window_size):

        self.meter = RunningStatistic(
            statistic_cls=MedianMeter,
            window_size=window_size)

    def fit_partial(self, score):
        """Fits the windowed postprocessor to the (next) timestep's score.

        Args:
            score (float): Input score.

        Returns:
            object: self.
        """
        self.meter.update(score)

        return self

    def transform_partial(self, score=None):
        """Applies postprocessing to the score using the window.

        Args:
            score (float): The input score.

        Returns:
            float: Transformed score.
        """
        return self.meter.get()


class RunningZScorePostprocessor(BasePostprocessor):
    """A postprocessor that normalizes score using Z-score normalization with the statistics of the window.

        Args:
            window_size (int): Length of the window
    """

    def __init__(self, window_size):

        self.variance_meter = RunningStatistic(
            statistic_cls=VarianceMeter, window_size=window_size)
        self.average_meter = RunningStatistic(
            statistic_cls=AverageMeter, window_size=window_size)

    def fit_partial(self, score):
        """Fits the windowed postprocessor to the (next) timestep's score.

        Args:
            score (float): Input score.

        Returns:
            object: self.
        """
        self.variance_meter.update(score)
        self.average_meter.update(score)

        return self

    def transform_partial(self, score):
        """Applies postprocessing to the score using the window.

        Args:
            score (float): The input score.

        Returns:
            float: Transformed score.
        """
        zscore = (score - self.average_meter.get()) / \
            np.sqrt(self.variance_meter.get())

        return zscore



from pysad.core.base_postprocessor import BasePostprocessor
import numpy as np

from pysad.utils.window import Window, UnlimitedWindow


class ConformalProbabilityCalibrator(BasePostprocessor):
    """This class provides an interface to convert the scores into probabilities through conformal prediction. Note that :cite:`laxhammar2013online` fits conformal calibration to already fitted samples' scores by the model whereas :cite:`ishimtsev2017conformal` fits the conformal calibration to some window of previous samples that are just before the target instance.
    This calibrator transforms by providing target score divided by the number of instances that are fitted before to this calibrator as transformation result.

        Args:
            windowed (bool): Whether the probability calibrator is windowed so that forget scores that are older than `window_size`.
            window_size (int): The size of window for running average and std. Ignored if `running_statistics` parameter is False.
    """

    def __init__(self, windowed=True, window_size=300):
        self.windowed = windowed
        self.window_size = window_size
        self.window = Window(window_size=self.window_size) if self.windowed else UnlimitedWindow()

    def fit_partial(self, score):
        """Fits particular (next) timestep's score to train the postprocessor.

        Args:
            score (float): Input score.
        Returns:
            object: self.
        """
        self.window.update(score)

        return self

    def transform_partial(self, score):
        """Transforms given score.

        Args:
            score (float): Input score.

        Returns:
            float: Processed score.
        """
        return (np.sum(np.array(self.window.get()) > score)) / (len(self.window.get()))





class Window:
    """Window to limit the instances in list and keep the size fixed when full.

    Args:
        window_size (int): The size of the window.
    """

    def __init__(self, window_size):
        self.window_size = window_size
        self.window = []

    def update(self, num):
        """Adds new item to the window. Removes the tail if size exceeds the self.window_size.

        Args:
            num (float): item to be added to the window.
        """
        self.window.append(num)
        if len(self.window) > self.window_size:
            self.window = self.window[1:]

    def get(self):
        """Method to obtain the window list.

        Returns:
            window (list): List containing the items in window
        """
        return self.window


class UnlimitedWindow(Window):
    """Unlimited window implemented for convenience. This class only provides a list with unlimited size. Note that this does not fit to the memory for streaming data.
    """

    def __init__(self):
        super().__init__(None)

    def update(self, num):
        """Adds new item to the window. Removes the tail if size exceeds the self.window_size.

        Args:
            num (float): item to be added to the window.
        """
        self.window.append(num)



from pysad.core.base_transformer import BaseTransformer


class IdentityScaler(BaseTransformer):
    """A scaler that does not modify the input, which is added for convenience.
    """

    def __init__(self):
        super().__init__(-1)

    def fit_partial(self, X):
        """Convenience method that does not modify the input or the scaler.

        Args:
            X (np.float64 array of shape (num_features,)): Input feature vector.

        Returns:
            object: The scaler.
        """
        return self

    def transform_partial(self, X):
        """Convenience method that does not modify the input.

        Args:
            X (np.float64 array of shape (num_features,)): Input feature vector.

        Returns:
            X (np.float64 array of shape (features,)): The exact same input feature vector.
        """
        return X



from pysad.core.base_model import BaseModel


class NullModel(BaseModel):
    """The model that returns `0.5` for all instances, which is added for testing and pipelining convenience purposes.
    """

    def __init__(self):
        self.labels = []

    def fit_partial(self, X, y=None):
        """This method is ignored. Added for convenience.

        Args:
            X: any
            y: any

        Returns:
            object: Returns the self.
        """
        return self

    def score_partial(self, X):
        """Directly returns 0.5.

        Args:
            X: any (Ignored)

        Returns:
            float: Equal to `0.5`.
        """
        return 0.5
